module findsourceslist_mod
!  VNY 16.12.2004
!  VNY 10.06.2005
! This module generates the source list detected in the input image
!
   use dal
   use sasdal
   use errorHandling
   use OmParameters_mod ! for computing the source moments

! VNY 16.12.2004
   use OmSourceParam_mod ! Source definitions

   implicit none

   private

   public FindSourcesList
!   public FindSourcesList2
!   public EstimateSourcePositions
!   public EstimateSourcePositions2
!   public CheckAllocation
!   public CheckDeallocation
   public SimpleImageSmoothing
   public AnalyseZeroOrder


contains

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! FindSourcesList
! Version for OMGRISM 04.06.2005
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subroutine FindSourcesList(image, xMin, xMax, yMin, yMax, ismoothWindow, &
     nSources, nSigma, nSourcesList, sourceList, crowdedField)
  ! Input and unmodified variables
  integer(int32), intent(in) :: crowdedField
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(inout) :: image(xMin:xMax, yMin:yMax)       
  ! Local variables
  integer(int32) :: i, j, k,l, i1,j1
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:, :) :: bkgImage
  real(single), allocatable, dimension(:, :), target :: bkgSubImage
  real(single), allocatable, dimension(:, :), target :: bkgSubImage2
  real(single), allocatable, dimension(:, :), target :: safeImage
  real(single), allocatable, dimension(:, :) :: maskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage2
  real(single), allocatable, dimension(:, :) :: imaskImage3
!  integer(int32), intent(in) :: nSourcesMax
!  real(single), intent(out) :: xPos(1:nSourcesMax)
!  real(single), intent(out) :: yPos(1:nSourcesMax)
!  real(single), intent(out) :: semiA(nSourcesMax)
!  real(single), intent(out) :: semiB(nSourcesMax)
!  real(single), intent(out) :: theta(nSourcesMax)
  integer(int32), intent(out) :: nSources
  real(single), intent(in) :: nSigma

! 16.12.2004
  integer(int32), intent(out) :: nSourcesList
  type(SourceLinkedList), intent(inout) :: sourceList 
  type(SourceParameters) :: srcParametersData

  real(single) :: sum1!,sum2,sum3
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: dim(2)
  integer(int32) :: idoubleWindow
  real(single), dimension(-ismoothWindow:ismoothWindow,-ismoothWindow:ismoothWindow ) :: template1
  real(single), dimension(-ismoothWindow:ismoothWindow,-2*ismoothWindow-1:2*ismoothWindow+1) :: template2
  real(single) :: umbral
!  real(single) :: r11,r12,r21,r22
  integer(int32) :: nPixels, nPixels0
  character(len=1024) :: regionFileName
  real(single) :: sumX, sumY
  real(single) :: sumValue, bkgValue
!  real(single) :: x0,y0
!  real(single) ::  x1, y1
  integer(int32) :: flag
!  integer(int32) ::radius
  integer(int32) :: reject
!  real(single) :: sqrMax, sqrMin, sqrAver
!  integer(int32) :: imaskValue
!  real(single) :: threshold, maskAver, maskMax
  real(single) :: sumValue2a, bkgValue2a!, bkgValue3a
!  integer(int32) :: ima,jma

!-- 13.01.05
  real(kind=single), pointer :: oswImage(:, :)
  real(kind=single), pointer :: bsoswImage(:, :)
  integer(kind=int16) :: sourceNumber
  integer(kind=int32) :: iMin, iMax, jMin, jMax
  real(kind=single) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 
!--

  integer(int32) :: nSourcesMax, verbFlag
  real(single) :: rMax, rMin, rAver, rAver2, rnAver, bkgFactor
  integer(int32) :: fillCrit, fillingFactor
  integer(int32) :: currentNpix
  bkgFactor=2.0
  nSourcesMax=20000
  !write(*,*) " FIND SOURCES LIST L103 crowdedField=", crowdedField, &
  !     " iSmoothWindow=", ismoothWindow

  regionFileName="sources.reg"

  template1=0.0
  do i=-ismoothWindow, ismoothWindow
     do j=ismoothWindow-1,ismoothWindow
        k=i*j
        if (k .lt. 0) template1(i,j)=-1.0
        if (k .gt. 0) template1(i,j)=1.0
     enddo
  enddo

  template2=0.0
  do j=ismoothWindow*2-1,ismoothWindow*2+1
     template2(:,j)=1.0
     template2(:,-j)=-1.0
  enddo

  do j=-2,2
     template2(:,j)=1.0
  enddo

  idoubleWindow=ismoothWindow*2+1

!  write(*,*) " L113 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
  dim(1) = lbound(image,1)
  dim(2)= ubound(image,1)
!  write(*,*) " L116 lbound 1 of image: dim(1)=", dim(1), "  dim(2)=", dim(2)
  dim(1) = lbound(image,2)
  dim(2)= ubound(image,2)
!  write(*,*) " L119 lbound 2 of image: dim(1)=", dim(1), "  dim(2)=", dim(2)
 ! write(*,*) " FIND SOURCES LIST L126"
  allocate (bkgImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgImage")  

  allocate (bkgSubImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgSubImage")  

  allocate (bkgSubImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgSubImage2")  

  allocate (safeImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "safeImage")  

  allocate (maskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  

  allocate (imaskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  

  allocate (imaskImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage2")  

  allocate (imaskImage3(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage3")  

  safeImage=image
  oswImage => safeImage

  ! shift the image by 3 pixels vertical
  do i=xMin, xMax
     do j=yMin,yMax-3
        image(i,j)=image(i,j)-bkgImage(i,j+3)
     enddo
  enddo
  ! shift the image by 3 pixels horizontal
  bkgImage=image
  do i=xMin, xMax-3
     do j=yMin,yMax
        image(i,j)=image(i,j)-bkgImage(i+3,j)
     enddo
  enddo

! imaskImage3 is stored to check it at the end of the programme
! NB
!  imaskImage3=image
! NB
  !write(*,*) " FIND SOURCES LIST L171"
  maskImage=0.0
!  umbral=10.0
  umbral=0.01


!  do i=xMin,xMax
!     do j=yMin,yMax!!
!        if (i-8 .ge. xMin .and. i+8 .le. xMax) then
!           if (j-8 .ge. yMin .and. j+8 .le. yMax) then
!              r11=image(i-8,j-8)
!              r12=image(i+7,j-8)
!              r21=image(i-8,j+7)
!              r22=image(i+7,j+7)
!              if (r21 .lt. -umbral &
!                   .and. r11 .gt. umbral &
!                   .and. r22 .gt. umbral &
!                   .and. r12 .lt. -umbral) then
!                 maskImage(i,j)=maskImage(i,j)+r11-r12-r21+r22
!              endif
!           endif
!        endif!
!
!     enddo
!  enddo
  !write(*,*) " FIND SOURCES LIST L196"
!------ VNY 04.06.05 Compute the mask as the contract of the image
  do i=xMin+8,xMax-8
     do j=yMin+8,yMax-8!!
        rMin=99999
        rMax=0
        if (safeImage(i,j) .gt. 0.0) then
           do i1=i-8,i+8
              do j1=j-8,j+8
                 if (image(i1,j1) .gt. rMax .and. safeImage(i1,j1) .gt. 0.0) rMax=image(i1,j1)
                 if (image(i1,j1) .lt. rMin .and. safeImage(i1,j1) .gt. 0.0) rMin=image(i1,j1)
              enddo
           enddo
           maskImage(i,j)=rMax-rMin
        endif
     enddo
  enddo


  rMin=99999.9
  rMax=0.0
  rAver=0.0
  rnAver=-.0
  do i=xMin+8,xMax-8
     do j=yMin+8,yMax-8!!
        if (safeImage(i,j) .gt. 0.0) then
           if (maskImage(i,j) .gt. rMax) rMax=maskImage(i,j)
           if (maskImage(i,j) .lt. rMin) rMin=maskImage(i,j)
           rAver=rAver+maskImage(i,j)
           rnAver=rnAver+1.0
        endif
     enddo
  enddo
  !write(*,*) " L220 sum=", rAver, " n=", rnAver
  rAver=rAver/rnAver
  !write(*,*) " L221 rMax=", rMax, " rMin=", rMin, " rAver=", rAver

  !write(*,*) " FIND SOURCES LIST L233"
  ! Compute rAver without bright sources
  rAver2=0.0
  rnAver=0.0
  bkgValue=0.0
  do i=xMin+8,xMax-8
     do j=yMin+8,yMax-8!!
        if (safeImage(i,j) .gt. 0.0 .and. maskImage(i,j) .lt. rAver) then
           rAver2=rAver2+maskImage(i,j)
           bkgValue=bkgValue+safeImage(i,j)
           rnAver=rnAver+1.0
        endif
     enddo
  enddo  
  rAver2=rAver2/rnAver
  bkgValue=bkgValue/rnAver
  !write(*,*) " L251 rAver2=", rAver2, " bkgValue=", bkgValue

  bkgSubImage=safeImage
  ! obtain the bkg-subtracted image
  if (crowdedField .gt. 0) then
     bkgFactor=10.0
  endif
  do i=xMin,xMax
     do j=yMin,yMax
        if (safeImage(i,j) .gt. 0.0) then
!           bkgSubImage(i,j)=bkgSubImage(i,j)-bkgValue
           bkgSubImage(i,j)=bkgSubImage(i,j)-bkgFactor*bkgValue
           if (bkgSubImage(i,j) .lt. 0.0) bkgSubImage(i,j) =0.0
        endif
     enddo
  enddo  
  bsoswImage => bkgSubImage2
  !write(*,*) " FIND SOURCES LIST L261"
  ! Clean the mask with 6*sigma criterium
  if (crowdedField .gt. 0) then
     where(maskImage .lt. real(crowdedField)*rAver) maskImage=0.0
  else
     where (maskImage .lt. rAver) maskImage=0.0
  endif
!---------------------------------------------
! joining those pixels that are close enough
  fillingFactor=(2*ismoothWindow+1)*(2*ismoothWindow+1)
  ! 09.12.2005
  fillCrit=int(0.8*real(fillingFactor)+0.5)
  if (crowdedField .gt. 0) then
     fillCrit=fillingFactor
  endif
  !
  imaskImage2=maskImage
  do i=xMin,xMax
     do j=yMin,yMax
        if (maskImage(i,j) .eq. 0) then
           sum1=0.0
           do k=i-ismoothWindow,i+ismoothWindow
              do l=j-ismoothWindow,j+ismoothWindow
                 if (k .ge. xMin .and. k .le. xMax) then
                    if (l .ge. yMin .and. l .le. yMax) then
                       if (maskImage(k,l) .gt. 0) then
                          sum1=sum1+1.0
                       endif
                    endif
                 endif
              enddo
           enddo
           if (crowdedField .gt. 0) then
              if (sum1 .gt. fillCrit) imaskImage2(i,j)=1
           else              
              if (sum1 .gt. fillCrit/2) imaskImage2(i,j)=1
           endif
        endif
     enddo
  enddo


! NB
!  imaskImage3=imaskImage2
!  imaskImage3=maskImage
! NB

!  call ComputeBackground(safeImage, maskImage, bkgImage, xMin,xMax,yMin,yMax, maskAver, maskMax)

! NB
!  imaskImage3=bkgImage
! NB

!  image=safeImage-bkgImage
!  where (image .lt. 0.0) image=0.0

!  sqrMax=0.0
!  sqrMin=1.0E25
!  sqrAver=0.0
!  sum3=0.0
!  do i=xMin, xMax
!     do j=yMin,yMax
!        sum2=sqrt(bkgImage(i,j))
!        if (sum2 .gt. sqrMax) sqrMax=sum2
!        if (sum2 .lt. sqrMin) sqrMin=sum2
!        sqrAver=sqrAver+sum2
!        sum3=sum3+1
!     enddo
!  enddo
!  if (sum3 .gt. 1.0) sqrAver=sqrAver/sum3
!  write (*,*) " L309 sqrMax=", sqrMax, " sqrMin=", sqrMin, " sqrAver=" , sqrAver


!  maskImage=image-bkgImage*sqrt(bkgImage)


  where (maskImage .lt. 0.0) maskImage=0.0

  imaskImage=imaskImage2

! NB
!  imaskImage3=maskImage
! NB

!  ! protect the high s/n sources with a negative boundary
!  imaskImage2=0
!  where (maskImage .gt. 0.0) imaskImage2=-5
!  imaskImage=0
!  do i=xMin,xMax
!     do j=yMin, yMax
!        if (imaskImage2(i,j) .lt. 0) then
!           imaskImage(i,j)=-5
!           if (i-1 .ge. xMin) imaskImage(i-1,j)=-5
!           if (i+1 .le. xMax) imaskImage(i+1,j)=-5
!           if (i-2 .ge. xMin) imaskImage(i-2,j)=-5
!           if (i+2 .le. xMax) imaskImage(i+2,j)=-5
!        endif
!     enddo
!  enddo
!  imaskImage2=imaskImage
!  do i=xMin,xMax
!     do j=yMin, yMax
!        if (imaskImage2(i,j) .lt. 0) then
!           imaskImage(i,j)=-5
!           if (j-1 .ge. yMin) imaskImage(i,j-1)=-5
!           if (j+1 .le. yMax) imaskImage(i,j+1)=-5
!           if (j-2 .ge. yMin) imaskImage(i,j-2)=-5
!           if (j+2 .le. yMax) imaskImage(i,j+2)=-5
!        endif
!     enddo
!  enddo
!  where(maskImage .gt. 0.0) imaskImage=1!
!
!!+++++++++++++++++++++++++++++++++++++++++++++++++++++=
!! second step
!  ! 
!  imaskValue=4
!  threshold=40.0 ! 35.0
!  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
!       yMin, yMax, imaskValue, threshold)
!
!!++++++++++++++++++++++++++++++++++++++++++++++++++++
!! third step
!  !
!  imaskValue=3
!  threshold=20.0 !8.0
!  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
!       yMin, yMax, imaskValue, threshold)
!  !
!!++++++++++++++++++++++++++++++++++++++++++++++++++++
!! fourth step
!  !
!  imaskValue=2
!  threshold=8.0 !4.0
!  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
!       yMin, yMax, imaskValue, threshold)
!!++++++++++++++++++++++++++++++++++++++++++++++++++++
!! fifth step
!  !
!  imaskValue=1
!  threshold= 3.5
!  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
!       yMin, yMax, imaskValue, threshold)
!
!!--- Test removing the line features
!!  call CheckLineFeatures(imaskImage, imaskImage2, xMin, xMax, yMin, yMax)
!!  imaskImage=imaskImage-imaskImage2
!!  !maskImage=real(imaskImage-imaskImage2)
!!  maskImage=real(imaskImage2)
!!---
!  maskImage=real(imaskImage)

!  imaskImage=0
!  where(maskImage .gt. 0.0) imaskImage=1
!  call CircleMaskSmoothing(imaskImage, imaskImage2, xMin, xMax, yMin, yMax, ismoothWindow)
!  imaskImage=imaskImage2
!  imaskImage3=imaskImage

!  !*******************************************************
!! gather the information about the sources
  !write(*,*) " FIND SOURCES LIST L408"
  nSources=0
  do i=xMin,xMax
     do j=yMin,yMax
        if (imaskImage(i,j) .gt. 0) then
           !write(*,*) "L413 i=", i, " j=", j
           nSources=nSources+1
           if (nSources .gt. nSourcesMax) nSources=nSourcesMax
           imaskImage(i,j)=-nSources
           nPixels=1 ! this will be the number of connected pixels
           sumX=real(i)
           sumY=real(j)
           sumValue=image(i,j)
           bkgValue=bkgImage(i,j)

           verbFlag=0
           !write(*,*) "L424"
!           ! Check if the pixel is connected to the others (recursive procedure)
!           nPixels0=CheckConnectedPixels(image, bkgImage, imaskImage,xMin,xMax,yMin,&
!                yMax,i,j,nSources, sumX, sumY, sumValue, bkgValue, verbFlag)
           currentNpix=0
           nPixels0=CheckConnectedPixels(bkgSubimage, bkgImage, imaskImage,xMin,xMax,yMin,&
                yMax,i,j,nSources, sumX, sumY, sumValue, bkgValue, verbFlag, &
                currentNpix)

           nPixels=nPixels+nPixels0
           !write(*,*) "L431 i=", i, " j=", j, " nPixels=", nPixels
           !xPos(nSources)=sumX+1.0
           !yPos(nSources)=sumY+1.0
           srcParametersData%xPos=sumX+1.0
           srcParametersData%yPos=sumY+1.0

           !semiA(nSources)=sqrt(real(nPixels))
           !semiB(nSources)=semiA(nSources)
           srcParametersData%semiA=sqrt(real(nPixels))
           srcParametersData%semiB=srcParametersData%semiA


           flag=0
           reject=0

           ! Reject the sources with small number of pixels or fainter than a threshold
           !write(*,*) " L333 nSigma=", nSigma
           if (nSigma .gt. 0.0) then
              if (nPixels .lt. 4) reject=1
           endif

           sumValue2a=sumValue
           bkgValue2a=bkgValue
           
!           if (nPixels .lt. 2) then
!              ima=int(sumX)
!              jma=int(sumY)
!              ! Check the pixel value and the background around
!              sumValue2a=0.0
!              if (ima .ge. xMin .and. ima .le. xMax .and. jma .ge. yMin .and. jma .le. yMax) &
!                   sumValue2a=safeImage(ima,jma)
!              ! background around
!              bkgValue2a=0.0
!              bkgValue3a=0.0
!              ima=i
!              jma=j+2
!              if (ima .ge. xMin .and. ima .le. xMax .and. jma .ge. yMin .and. jma .le. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              jma=j-2
!              if (ima .ge. xMin .and. ima .le. xMax .and. jma .ge. yMin .and. jma .le. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              jma=j
!              ima=i-2
!              if (ima .ge. xMin .and. ima .le. xMax .and. jma .ge. yMin .and. jma .le. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              ima=i+2
!              if (ima .ge. xMin .and. ima .le. xMax .and. jma .ge. yMin .and. jma .le. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              ima=i-2
!              jma=j-2
!              if (ima .gt. xMin .and. ima .lt. xMax .and. jma .gt. yMin .and. jma .lt. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              jma=j+2
!              if (ima .gt. xMin .and. ima .lt. xMax .and. jma .gt. yMin .and. jma .lt. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              ima=i+2
!              jma=j-2
!              if (ima .gt. xMin .and. ima .lt. xMax .and. jma .gt. yMin .and. jma .lt. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!              jma=j+2
!              if (ima .gt. xMin .and. ima .lt. xMax .and. jma .gt. yMin .and. jma .lt. yMax) then
!                 bkgValue2a=bkgValue2a+safeImage(ima,jma)
!                 bkgValue3a=bkgValue3a+1.0
!              endif
!
!              if (bkgValue3a .gt. 1.0) bkgValue2a=bkgValue2a/bkgValue3a
!              if (sumValue2a .le. bkgValue2a+3.0*sqrt(bkgValue2a)) reject=1
!           endif

! Test printing into a file
           !write(8,*) " Source=", nSources, " i=", i+1, " j=", j+1, &
           !     " sumX=", sumX, " sumY=", sumY, " nPixels=", nPixels,  &
           !     " v2=", sumValue2a, " b2=", &
           !     bkgValue2a+6.0*sqrt(bkgValue2a), " flag=", flag, " reject=", reject


           if (reject .gt. 0) then
              nSources=nSources-1
           else
              ! 16.12.2004
              srcParametersData%number=nSources
              srcParametersData%xPos=sumX+1.0
              srcParametersData%yPos=sumY+1.0
              srcParametersData%semiA=sqrt(real(nPixels))
              srcParametersData%semiB=srcParametersData%semiA
              srcParametersData%theta=0.0
              srcParametersData%noSourcePixels=nPixels
              srcParametersData%rawBackGroundCountRate=bkgValue2a
              srcParametersData%rawBackGroundCountRateErr=sqrt(bkgValue2a)
              srcParametersData%rawCountRate=sumValue2a
              srcParametersData%rawCountRateErr=sqrt(sumValue2a)
              srcParametersData%iMin=max(xMin, int(sumX+1.0-srcParametersData%semiA))
              srcParametersData%iMax=min(xMax, int(sumX+1.0+srcParametersData%semiA))
              srcParametersData%jMin=max(yMin, int(sumY+1.0-srcParametersData%semiB))
              srcParametersData%jMax=min(yMax, int(sumY+1.0+srcParametersData%semiB))
!              call AddSourceToList(sourceList, srcParametersData)
              
              !--- 13.01.05 
              !if (nSources .eq. 1) then
              sourceNumber=nSources
              iMin=srcParametersData%iMin
              iMax=srcParametersData%iMax
              jMin=srcParametersData%jMin
              jMax=srcParametersData%jMax
              !call ComputeMoments(sourceNumber, oswImage, iMin, iMax, &
              !     jMin, jMax, xPos, yPos, xPosErr, yPosErr, aValue, bValue, &
              !     theta)
              bsoswImage=0.0
              where(imaskImage .eq. -nSources) bsoswImage=bkgSubImage
              xPos=-1.0
              yPos=-1.0
              xPosErr=-1.0
              yPosErr=-1.0
              aValue=-1.0
              bValue=-1.0
              theta=-1.0
              call ComputeMoments(sourceNumber, bsoswImage, iMin, iMax, &
                   jMin, jMax, xPos, yPos, xPosErr, yPosErr, aValue, bValue, &
                   theta)
              !write(*,*) "L549************************************** nSource=", nSources
              !write (*,*) " initinal xPos=", srcParametersData%xPos, &
              !     " yPos=", srcParametersData%yPos
              !write(*,*) " iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax
              !write(*,*) " xPos=", xPos, " yPos=", yPos, " xPosErr=", xPosErr, &
              !     " yPosErr=", yPosErr, " aValue=", aValue, " bValue=", bValue, &
              !     " theta=", theta
              ! 06.09.2005 Assign the moment values to the source parameters
              if (xPos .gt. iMin .and. xPos .lt. iMax &
                   .and. yPos .gt. jMin .and. yPos .lt. jMax) then
                 srcParametersData%xPos=xPos
                 srcParametersData%yPos=yPos
                 srcParametersData%semiA=aValue
                 srcParametersData%semiB=bValue
                 srcParametersData%theta=theta*radiansToDegrees
              endif
              !write(*,*) "L593 x_new=", srcParametersData%xPos, " y_ne=", srcParametersData%yPos, &
              !     " semiA=", srcParametersData%semiA, " semiB=", srcParametersData%semiB, &
              !     " theta=", srcParametersData%theta
              call AddSourceToList(sourceList, srcParametersData)
              !endif
              !---
           endif
        endif ! (maskImage
     enddo ! j=
  enddo ! i=
  !write(*,*) " FIND SOURCES LIST L563"
  ! 16.12.2004
  nSourcesList=nSources

  !write(*,*) "**************************************************"
  !write(*,*) " L587 Total number of sources=", nSources
!! write(*,*) " maskAver=", maskAver, " maskMax=", maskMax, " 26*sqrt*maskAver=", 26.0*sqrt(maskAver)

! 16.12.2004
!  call WriteSourceRegion(regionFileName, sourceList, nSourcesList)
!  write(*,*) "L592  Region2 file written"

!! This image will be passed to the output for checking
  image=imaskImage
!!  image=imaskImage3 
!!  image=safeImage
!  write(*,*) " FIND SOURCES LIST L579"
!*************************************************
  if(allocated(bkgImage)) then
     deallocate(bkgImage,stat=istat)
     call CheckDeallocation(allocErr, "bkgImage")
  end if

  if(allocated(bkgSubImage)) then
     deallocate(bkgSubImage,stat=istat)
     call CheckDeallocation(allocErr, "bkgSubImage")
  end if

  if(allocated(bkgSubImage2)) then
     deallocate(bkgSubImage2,stat=istat)
     call CheckDeallocation(allocErr, "bkgSubImage2")
  end if

  if(allocated(imaskImage)) then
     deallocate(imaskImage,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage")
  end if

  if(allocated(imaskImage2)) then
     deallocate(imaskImage2,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage2")
  end if

  if(allocated(imaskImage3)) then
     deallocate(imaskImage3,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage3")
  end if

  if(allocated(maskImage)) then
     deallocate(maskImage,stat=istat)
     call CheckDeallocation(allocErr, "maskImage")
  end if

  if(allocated(safeImage)) then
     deallocate(safeImage,stat=istat)
     call CheckDeallocation(allocErr, "safeImage")
  end if

end subroutine FindSourcesList


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! AnalyseZeroOrder (actually, recomputing the position of the zero-order)
! Version for OMGRISM 06.09.2005
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subroutine AnalyseZeroOrder(image, patternImage, xMin, xMax, yMin, yMax, &
     iMin,iMax,jMin,jMax,xPos,yPos,xPosErr,yPosErr,aValue,bValue,theta )
  ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax)
  real(single), intent(in) :: patternImage(xMin:xMax,yMin:yMax)
  integer(kind=int32), intent(in) :: iMin, iMax, jMin, jMax
  real(kind=single),intent(out) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 

  ! Local variables
  real(single), allocatable, dimension(:, :) :: bkgImage
  real(single), allocatable, dimension(:, :), target :: bkgSubImage
  real(single), allocatable, dimension(:, :), target :: bkgSubImage2
  real(single) :: rnAver
  real(single) :: bkgValue
  integer(kind=int16) :: sourceNumber
  integer(int32) :: i, j, ic,jc
  integer(int32) :: ipMin, ipMax, jpMin, jpMax, imaskValue

  integer(int32) :: allocErr   
  integer           :: istat

  real(kind=single), pointer :: bsoswImage(:, :)



!-------------------------------------
  ! tests 07.09.2005
  ic=(iMin+iMax)/2
  jc=(jMin+jMax)/2
  iMaskValue=patternImage(ic,jc)
  !write(*,*) "L700  iMaskValue=", iMaskValue

  xPos=ic
  yPos=jc
  xPosErr=0.0
  yPosErr=0.0
  aValue=0.0
  bValue=0.0
  theta=0.0

  ! Compute the moment centroid coordinates only if there is a source there
  if (iMaskValue .lt. 0) then 
     allocate (bkgImage(xMin : xMax, yMin : yMax), stat = allocErr)
     call CheckAllocation(allocErr, "bkgImage")  

     allocate (bkgSubImage(xMin : xMax, yMin : yMax), stat = allocErr)
     call CheckAllocation(allocErr, "bkgSubImage")  
     
     allocate (bkgSubImage2(xMin : xMax, yMin : yMax), stat = allocErr)
     call CheckAllocation(allocErr, "bkgSubImage2")  

     ipMin=iMin
     ipMax=iMax
     jpMin=jMin
     jpMax=jMax

     ! Find the real sizes of the source 
     do i=ic,xMin+1,-1
        if (patternImage(i,jc) .lt. 0 .and. patternImage(i-1,jc) .ge. 0) then
           ipMin=i-1
           exit
        endif
     enddo
     do i=ic,xMax-1
        if (patternImage(i,jc) .lt. 0 .and. patternImage(i+1,jc) .ge. 0) then
           ipMax=i+1
           exit
        endif
     enddo

     do j=jc,yMin+1,-1
        if (patternImage(ic,j) .lt. 0 .and. patternImage(ic,j-1) .ge. 0) then
           jpMin=j-1
           exit
        endif
     enddo
     do j=jc,yMax-1
        if (patternImage(ic,j) .lt. 0 .and. patternImage(ic,j+1) .ge. 0) then
           jpMax=j+1
           exit
        endif
     enddo
     !write(*,*) " L743 ipMin=", ipMin, " ipMax=", ipMax, " jpMin=", jpMin, " jpMax=", jpMax


     ! Compute bkgValue without bright sources
     rnAver=0.0
     bkgValue=0.0
     do i=xMin+8,xMax-8
        do j=yMin+8,yMax-8!!
           if (image(i,j) .gt. 0.0 .and. patternImage(i,j) .eq. 0) then
              bkgValue=bkgValue+image(i,j)
              rnAver=rnAver+1.0
           endif
        enddo
     enddo
     if (rnAver .gt. 1.0) then
        bkgValue=bkgValue/rnAver
     endif
     !write(*,*) " L779 bkgValue=", bkgValue

     bkgSubImage=image
     ! obtain the bkg-subtracted image
     do i=xMin,xMax
        do j=yMin,yMax
           if (image(i,j) .gt. 0.0) then
              bkgSubImage(i,j)=bkgSubImage(i,j)-2.0*bkgValue
              if (bkgSubImage(i,j) .lt. 0.0) bkgSubImage(i,j) =0.0
           endif
        enddo
     enddo

     bsoswImage => bkgSubImage2

     sourceNumber=-iMaskValue
     bsoswImage=0.0
     where(patternImage .eq. iMaskValue) bsoswImage=bkgSubImage
     call ComputeMoments(sourceNumber, bsoswImage, ipMin, ipMax, &
          jpMin, jpMax, xPos, yPos, xPosErr, yPosErr, aValue, bValue, &
          theta)

     !write(*,*) "L782  ipMin=", ipMin, " ipMax=", ipMax, " jpMin=", jpMin, " jpMax=", jpMax
     !write(*,*) " xPos=", xPos, " yPos=", yPos, " xPosErr=", xPosErr, &
     !     " yPosErr=", yPosErr, " aValue=", aValue, " bValue=", bValue, &
     !     " theta=", theta

     if(allocated(bkgImage)) then
        deallocate(bkgImage,stat=istat)
        call CheckDeallocation(allocErr, "bkgImage")
     end if

     if(allocated(bkgSubImage)) then
        deallocate(bkgSubImage,stat=istat)
        call CheckDeallocation(allocErr, "bkgSubImage")
     end if

     if(allocated(bkgSubImage2)) then
        deallocate(bkgSubImage2,stat=istat)
        call CheckDeallocation(allocErr, "bkgSubImage2")
     end if
  endif ! iMaskValue .lt. 0

end subroutine AnalyseZeroOrder


!****************************************************************
! Subroutine to write the source region
!****************************************************************
subroutine WriteSourceRegion(regionFileName, sourceList, nSourcesList)

!  character(len=1024), intent(in) :: regionFileName
  character(len=255), intent(in) :: regionFileName
  integer(int32), parameter :: ioUnit=2
  integer(int32) :: i

! 16.12.2004
  integer(int32), intent(out) :: nSourcesList
  type(SourceLinkedList), intent(inout) :: sourceList 
  type(SourceParameters), pointer :: sourceData(:) => null()

  if (nSourcesList > 0) then
     sourceData => SourceDataArray(sourceList, nSourcesList)

  endif

  open(unit = ioUnit, file = regionFileName, status = 'unknown')
  do i = 1, nSourcesList
     write(ioUnit, '('' ELLIPSE('', f7.2,'', '', f7.2, '', '', f6.2, '', '', f6.2, '', '', f6.2, '')'')') &
!              xPos(i), yPos(i), semiA(i), semiB(i), theta(i)
              sourceData(i)%xPos, sourceData(i)%yPos, sourceData(i)%semiA, sourceData(i)%semiB, sourceData(i)%theta
    end do
    close(ioUnit)
  !
end subroutine WriteSourceRegion



!****************************************************************
! Subroutine for delimiting boundaries around the sources
!****************************************************************
subroutine SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, yMin, yMax, &
     imaskValue, threshold)

  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax) 
  real(single), intent(in) :: bkgImage(xMin:xMax, yMin:yMax) 
  integer(int32), intent(inout) :: imaskImage(xMin:xMax, yMin:yMax)       
  integer(int32), intent(in) :: imaskValue
  real(single), intent(in) :: threshold

  real(single), allocatable, dimension(:, :) :: maskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage2
  integer(int32), allocatable, dimension(:, :) :: imaskImage3
  integer(int32) :: i,j
  integer(int32) :: allocErr   
  integer           :: istat

  allocate (maskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  
  allocate (imaskImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage2")  
  allocate (imaskImage3(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage3")  

! The threshold parameter can cut off many faint sources (in case  it is needed)
  maskImage=image-threshold*sqrt(bkgImage)
  !
  where (maskImage .gt. 0.0) imaskImage2=-3
  imaskImage3=0
  do i=xMin,xMax
     do j=yMin, yMax
        if (imaskImage2(i,j) .lt. 0) then
           imaskImage3(i,j)=-imaskValue
           if (i-1 .ge. xMin) imaskImage3(i-1,j)=-imaskValue
           if (i+1 .le. xMax) imaskImage3(i+1,j)=-imaskValue
        endif
     enddo
  enddo
  imaskImage2=imaskImage3
  do i=xMin,xMax
     do j=yMin, yMax
        if (imaskImage2(i,j) .lt. 0) then
           imaskImage3(i,j)=-imaskValue
           if (j-1 .ge. yMin) imaskImage3(i,j-1)=-imaskValue
           if (j+1 .le. yMax) imaskImage3(i,j+1)=-imaskValue
        endif
     enddo
  enddo
  where (maskImage .gt. 0.0) imaskImage3=1
  imaskImage=imaskImage+imaskImage3

  if(allocated(imaskImage2)) then
     deallocate(imaskImage2,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage2")
  end if
  if(allocated(imaskImage3)) then
     deallocate(imaskImage3,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage3")
  end if
  if(allocated(maskImage)) then
     deallocate(maskImage,stat=istat)
     call CheckDeallocation(allocErr, "maskImage")
  end if
  !
end subroutine SourceBoundary


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! CheckConnectedPixels
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
recursive function  CheckConnectedPixels &
     (image, bkgImage, imaskImage, xMin, xMax, yMin, yMax, i,j, label, &
     sumX, sumY, sumValue, bkgValue, verbFlag, currentNpix) result(nPixels)
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax)
  real(single), intent(in) :: bkgImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(inout) :: imaskImage(xMin:xMax, yMin:yMax)       
  integer(int32), intent(in) :: i,j
  integer(int32), intent(in) :: label
  real(single), intent(inout) :: sumX, sumY, sumValue, bkgValue
  integer(int32), intent(inout) :: currentNpix
  integer(int32) :: k,l
  integer(int32) :: nPixels
  integer(int32) :: nPixels0, nPixels1
  integer(int32) :: verbFlag
  nPixels0=0
  if (currentNpix .gt. 10000) then
!  if (currentNpix .gt. 30000) then
     nPixels=0
     return
  endif
  !write(*,*) "Check connected pixels L1809 i=", i, " j=", j, " label=", label
  do k=i-1,i+1
     do l=j-1,j+1
        if (k .ge. xMin .and. k .le. xMax &
             .and. l .ge. yMin .and. l .le.yMax) then
           if (imaskImage(k,l) .gt. 0) then
              nPixels0=nPixels0+1
              currentNpix=currentNpix+1
              !write(*,*) "L1815 k=", k, " l=", l, " currentNpix=", currentNpix
              imaskImage(k,l)=-label
              !if (verbFlag .eq. 1) write(*,*) "L1714 k=", k, " l=", l
                 if (image(k,l) .gt. sumValue) then
                    sumValue=image(k,l)
                    bkgValue=bkgImage(k,l)
                    sumX=real(k)
                    sumY=real(l)
                 endif
                 !write(*,*) " L1825"
              nPixels1=CheckConnectedPixels(image, bkgImage, &
                   imaskImage,xMin,xMax,yMin,yMax,k,&
                   l,label, sumX, sumY, sumValue, bkgValue, verbFlag, &
                   currentNpix)
              nPixels0=nPixels0+nPixels1
              !write(*,*) "L1830 nPixels0=", nPixels0
           endif
        endif
     enddo ! l=
  enddo ! k=
  nPixels=nPixels0
end function CheckConnectedPixels



!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! ComputeBackground
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine ComputeBackground(inputImage, maskImage, bkgImage, xMin, xMax, yMin, yMax, &
     maskAver, maskMax)
  ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(in) :: maskImage(xMin:xMax, yMin:yMax)
  real(single), intent(inout) :: bkgImage(xMin:xMax, yMin:yMax)     
  real(single), intent(out) :: maskAver
  real(single), intent(out) :: maskMax
  ! Local variables
  integer(int32) :: i, j, k,l
  integer(int32) :: dim(2)
!  real(single) :: sum1
  real(single) :: sum2
  integer(int32) :: ismoothWindow, maskWindow
  real(single) :: threshold
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize!, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  real(single), allocatable, dimension(:, :) :: tmpImage

!---------------------------------------------
 
  ismoothWindow=4
  vectSize=ismoothWindow*16+4

  allocate (tmpImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "tmpImage")  

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

  dim(1) = lbound(inputImage,1)
  dim(2)= ubound(inputImage,1)
  !write(*,*) " L352 lbound 1 of inputImage: dim(1)=", dim(1), "  dim(2)=", dim(2)
  dim(1) = lbound(inputImage,2)
  dim(2)= ubound(inputImage,2)
  !write(*,*) " L355 lbound 2 of inputImage: dim(1)=", dim(1), "  dim(2)=", dim(2)
  !write(*,*) " mod(7,2)=", mod(7,2), " mod(8,2)=", mod(8,2)
  !write(*,*) " modulo(7,2)=", modulo(7,2), " modulo(8,2)=", modulo(8,2)

  !tmpImage=inputImage

! find average and maximal values of the mask
  maskAver=0.0
  maskMax=0.0
  do i=xMin,xMax
     do j=yMin,yMax
        if (maskImage(i,j) .gt. 0.0) then
           maskAver=maskAver+maskImage(i,j)
           sum2=sum2+1.0
           if (maskImage(i,j) .gt. maskMax) maskMax=maskImage(i,j)
        endif
     enddo
  enddo
  if (sum2 .gt. 1.0) maskAver=maskAver/sum2
  threshold=maskAver+0.3*(maskMax-maskAver)
  !write(*,*) " L377 maskAver=", maskAver, " maskMax=", maskMax, " threshold=", threshold, " <-----"

  ! simple smoothing of the input image
  call SimpleImageSmoothing(inputImage, tmpImage, xMin, xMax, yMin, yMax, ismoothWindow)
 
! applying the mask to the smoothed image
!  maskWindow=ismoothWindow/2
  maskWindow=ismoothWindow*8
  do i=xMin,xMax
     do j=yMin,yMax!!
        if (maskImage(i,j) .gt. 0.0) then
           !if (maskImage(i,j) .lt. maskAver) then
              do k=i-maskWindow+2,i+maskWindow+2
                 do l=j-maskWindow+2,j+maskWindow+2
                    if (k .ge. xMin .and. k .le. xMax) then
                       if (l .ge. yMin .and. l .le. yMax) then
                          tmpImage(k,l)=0.0
                       endif
                    endif
                 enddo
              enddo
           !else
           !   !if (maskImage(i,j) .lt. maskAver+0.5*maskMax) then
           !      do k=i-ismoothWindow+2,i+ismoothWindow+2
           !         do l=j-ismoothWindow+2,j+ismoothWindow+2
           !            if (k .ge. xMin .and. k .le. xMax) then
           !               if (l .ge. yMin .and. l .le. yMax) then
           !                  tmpImage(k,l)=0.0
           !               endif
           !            endif
           !         enddo
           !      enddo
           !   !endif
           !endif
        endif
     enddo
  enddo

! median smoothing of the masked image
  call MedianImageSmoothing(tmpImage, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(bkgImage, tmpImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(tmpImage, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(bkgImage, tmpImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call SimpleImageSmoothing(tmpImage, bkgImage, xMin, xMax, yMin, yMax, ismoothWindow)


  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if
  if(allocated(tmpImage)) then
     deallocate(tmpImage,stat=istat)
     call CheckDeallocation(allocErr, "tmpImage")
  end if

end subroutine ComputeBackground


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! SimpleImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine SimpleImageSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
     
  ! Local variables
  integer(int32) :: i, j, k,l
  real(single) :: sum1,sum2

!---------------------------------------------
  ! simple smoothing of the input image
  do i=xMin,xMax
     do j=yMin,yMax!!
        sum1=0.0
        sum2=0.0
        do k=i-ismoothWindow,i+ismoothWindow
           do l=j-ismoothWindow,j+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 if (l .ge. yMin .and. l .le. yMax) then
                    sum1=sum1+inputImage(k,l)
                    sum2=sum2+1.0
                 endif
              endif
           enddo
        enddo
        if (sum2 .gt. 1.0) sum1=sum1/sum2
        outputImage(i,j)=sum1
     enddo
  enddo
end subroutine SimpleImageSmoothing


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! MedianImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine MedianImageSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: i, j, k,l
  real(single) :: sum1

!---------------------------------------------
  vectSize=ismoothWindow*16+4

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

! median smoothing of the masked image
  outputImage=0.0
  do i=xMin, xmax
     do j=yMin,yMax
        ! fill in the vector (cross-shaped aperture)
        vect=0.0
        nElem=0
        l=j
        do k=i-ismoothWindow*2,i+ismoothWindow*2
           if (k .ge. xMin .and. k .le. xMax) then
              if (inputImage(k,l) .gt. 0.0) then
                 vect(nElem)=inputImage(k,l)
                 nElem=nElem+1
              endif
           endif
        enddo
        k=i
        do l=j-ismoothWindow*2,j+ismoothWindow*2
           if (l .ge. yMin .and. l .le. yMax) then
              if (inputImage(k,l) .gt. 0.0) then
                 vect(nElem)=inputImage(k,l)
                 nElem=nElem+1
              endif
           endif
        enddo
        if (nElem .gt. 0) then
           sum1=smoothMedian(vect,vectSize,nElem)
           outputImage(i,j)=sum1
        endif
     enddo
  enddo

  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if

end subroutine MedianImageSmoothing


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Median smoothing of a vector
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
real function smoothMedian(vect, vectSize, nmax)
  integer(int32), intent(in) :: vectSize
  integer(int32), intent(in) :: nmax ! number of significant elements in the vector, including 0-th
  real(single), intent(in) :: vect(0:vectSize-1)
  real(kind=double) :: a(0:vectSize-1) ! auxiliar vector
  integer :: k,l
  real(single) :: dum
  if (nmax .gt. 2) then
     ! Copy the vector
     a=vect
     ! ordering the vector a
     do k=0,nmax-2
        do l=k+1, nmax-1
           if (a(l) .lt. a(k)) then
              dum=a(k)
              a(k)=a(l)
              a(l)=dum
           endif
        enddo
     enddo
     if (mod(nmax,2) .eq. 0) then
        l=nmax/2
        smoothMedian=(a(l-1)+a(l))/2.0
     else
        l=(nmax-2)/2
        smoothMedian=a(l)
     endif
  else
     smoothMedian=0.0
     do k=0,nmax-1
        smoothMedian=smoothMedian+vect(k)
        dum=dum+1.0
     enddo
     if (dum .gt. 1.0) smoothMedian=smoothMedian/dum
  endif
end function smoothMedian

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine CircleMaskSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  integer(int32), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  integer(int32), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
     
  ! Local variables
  integer(int32) :: i, j, k,l
  real(single) :: sum1,sum2!, sum3
  real(single) :: distance!, meanValue

!---------------------------------------------
  ! simple smoothing of the input image
  do i=xMin,xMax
     do j=yMin,yMax!!
        sum1=0.0
        sum2=0.0

        do k=i-ismoothWindow,i+ismoothWindow
           do l=j-ismoothWindow,j+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 if (l .ge. yMin .and. l .le. yMax) then
                    distance=sqrt(real((i-k)*(i-k)+(j-l)*(j-l)))
                    if (distance .lt. ismoothWindow) then
                       sum1=sum1+inputImage(k,l)
                       sum2=sum2+1.0
                    endif
                 endif
              endif
           enddo
        enddo
        if (sum1 .gt. 10.0) then
           do k=i-2,i+2
              do l=j-2,j+2
                 if (k .ge. xMin .and. k .le. xMax) then
                    if (l .ge. yMin .and. l .le. yMax) then
                       outputImage(k,l)=1       
                    endif
                 endif
              enddo
           enddo
        endif
     enddo
  enddo
end subroutine CircleMaskSmoothing


!***************************************************
! CheckAllocation
!***************************************************
!subroutine CheckAllocation(status, string)
!    integer, intent(in) ::  status
!    character(len=*), intent(in) :: string
!    character(len=100) :: comment
!   
!    if(status .ne. 0) then
!       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
!       write (errstr, *) comment
!       call fatal("allocErr") 
!    end if
!!    noAllocations = noAllocations + 1
!end subroutine CheckAllocation
!!***************************************************
!! CheckDeallocation
!!***************************************************
!subroutine CheckDeallocation(status, string)
!    integer, intent(in) ::  status
!    character(len=*), intent(in) :: string
!    character(len=100) :: comment
!   
!    if(status .ne. 0) then
!       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
!       write (errstr, *) comment
!       call warning("allocErr", comment) 
!    end if
!!    noDeallocations = noDeallocations + 1
!end subroutine CheckDeallocation

!*********************************************
! CreateRegionFile
!*********************************************
!subroutine CreateRegionFile2(nSources, sourceData, regionFile)
!    integer(int32), intent(in) :: nSources!
!
!    type(sourceParameters), pointer :: sourceData(:)!
!
!    character(*), intent(in) :: regionFile
!    integer(int32) :: i
!    integer(int32), parameter :: ioUnit = 2
! 
!    
!    open(unit = ioUnit, file = regionfile, status = 'unknown')
!    do i = 1, nSources
!        write(ioUnit, '('' ELLIPSE('', f7.2,'', '', f7.2, '', '', f6.2, '', '', f6.2, '', '', f6.2, '')'')') &
!              sourceData(i)%xPos, sourceData(i)%yPos,   & 
!              sourceData(i)%semiA * 3, &    ! / (binx * plateScale), &
!              sourceData(i)%semiB * 3, &   ! / (biny * plateScale), 
!              sourceData(i)%theta
!    end do
!    close(ioUnit)
!    
!end subroutine CreateRegionFile2

end module findsourceslist_mod

